@@ -1,6 +1,6 @@
 #-----------------------------------------------------------------------------
 #
-#	$Id : Meta.pm 2.007 2005-09-12 JMG$
+#	$Id : Meta.pm 2.008 2006-03-22 JMG$
 #
 #	Initial developer: Jean-Marie Gouarne
 #	Copyright 2005 by Genicorp, S.A. (www.genicorp.com)
@@ -12,9 +12,9 @@
 
 package	OpenOffice::OODoc::Meta;
 use	5.006_001;
-our	$VERSION	= 2.007;
+our	$VERSION	= 2.008;
 
-use	OpenOffice::OODoc::XPath	2.204;
+use	OpenOffice::OODoc::XPath	2.215;
 require Exporter;
 our	@ISA		= qw ( OpenOffice::OODoc::XPath Exporter );
 our	@EXPORT		= qw ( ooLocaltime ooTimelocal );
@@ -74,6 +74,15 @@ sub	new
 		}
 	}
 
+#-----------------------------------------------------------------------------
+# overrides setText() because meta elements contain flat text only
+
+sub	setText
+	{
+	my $self	= shift;
+	return $self->setFlatText(@_);
+	}
+
 #-----------------------------------------------------------------------------
 # generic read/write accessor for text elements
 
